import json
import pandas as pd
import os
import numpy as np
from transformers import ViTFeatureExtractor
from torch.utils.data import Dataset
from torchvision.io import read_image


class IsingData(Dataset):
    '''
    This class serves as the PyTorch dataset of the processed images from `data_processer.py` script. Because stratified sampling
    is leveraged within the `DataProcessor`, this class loads the train, test, and validation sets immediately.

    params:
        - experiment_json_path (str): path to `experiments.json`. Will be found in `data_dir`.
        - train_test_validation (str): takes in either 'train', 'test', or 'validation' and will parse out
        the respective dataset specified. if left as default, None, it will return the entire set.
        - feature_extractor: take in HuggingFace's ViTFeatureExtractor to perform necessary transformations on
        images. Is also used for ResNet models as well. Currently, the datasets only support using ViTFeatureExtractor,
        so if you change this parameter, it will break this class.
    '''
    def __init__(
        self,
        experiment_json_path:str,
        train_test_validation:str=None,
        feature_extractor = ViTFeatureExtractor()
        ):
        super().__init__()
        
        # Construct set names for parsing out set
        self.__setnames__ = np.array(['train', 'test', 'validation'])

        # Instantiate path and load experiments json
        self.experiment_json_path = experiment_json_path
        with open(self.experiment_json_path, 'r') as j:
            self.experiments = json.load(j)
        
        # Parse out keys from experiments json
        self.data_dir = self.experiments['data_dir']
        self.num_labels = self.experiments['num_labels']
        self.intervals = np.array(self.experiments['intervals'])

        # Construct path to csv and load:
        self.csv_path = os.path.join(self.data_dir, 'csvs', 'data.csv')
        self.csv = pd.read_csv(self.csv_path)
        
        # Instantiate feature extractor
        self.feature_extractor = feature_extractor
        
        # Parse out set:
        self.train_test_validation = train_test_validation
        if self.train_test_validation:
            self._parse_set()
        

    def __len__(self):
        return len(self.csv)

    def __getitem__(self, idx):
        # Parse image from csv and load image
        img_path = self.csv.iloc[idx, 0]
        img = read_image(img_path)
        
        # Parse label from csv
        label = self.csv.iloc[idx, 1]
        
        # If the feature_extractor exists, use it on the image:
        if self.feature_extractor:
            img = self.feature_extractor(img, return_tensors='pt')['pixel_values']
        
        # Now return everything
        return img, label
    
    def _parse_set(self):
        '''
        Parse set_label column of `self.csv` to get train/test/validation set.
        '''
        try:
            # 0 if train, 1 if test, 2 if validation
            set_num = int(np.where(self.__setnames__ == self.train_test_validation)[0])
            
            # Now only select the part of the dataframe that is the set.
            self.csv = self.csv[self.csv['set_label'] == set_num].reset_index(drop=True) # Reset index for self.__getitem__()
        except:
            KeyError, 'The only keys allowed for train_test_valdiation is "train", "test", "validation", or None.'


import torch
import torch.nn as nn
import os
import numpy as np
import matplotlib.pyplot as plt
from torch.utils.data import DataLoader
from torch.optim import Optimizer
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay

# TODO: Create Trainer base class with all params except save_confusion_matrix
class ClassificationTrainer:
    '''
    Trainer class to handle all training, test, and evaluation functionality.
    
    - model (nn.Module): pytorch model architecture to be trained. must be a classification
    model.
    - model_name (str): name of model. will be used in naming the checkpoint.
    - train_dataloader (DataLoader): dataloader containing the training set.
    - test_dataloader (DataLoader): dataloader containing the test set.
    - loss_function (torch loss): loss function for training
    - optimizer (Optimizer): torch.optim Optimizer to use for optimization.
    - learning_rate (float): learning rate for training
    - batch_size (int): batch size of data.
    - num_epochs (int): number of epochs to train.
    - device (torch.device): set device for training. default set to torch.default("cpu").
    - display_step_size (int): how many batches to train on.
    - evaluate (bool): if true, evaluate on validation set.
    - save_confusion_matrix (bool): if true, save confusion matrix.
    '''
    def __init__(
        self, 
        model:nn.Module,
        model_name:str,
        train_dataloader:DataLoader,
        test_dataloader: DataLoader,
        loss_function:nn.Module,
        optimizer:Optimizer,
        learning_rate:float,
        batch_size:int,
        num_epochs:int,
        device:torch.device = torch.device('cpu'),
        display_step_size:int = 100,
        evaluate:bool = True,
        save_confusion_matrix:bool = True
        ):
        self.device = device
        self.model = model.to(self.device)
        self.model_name = model_name
        self.train_dataloader = train_dataloader
        self.test_dataloader = test_dataloader
        
        self.test_size = len(self.test_dataloader.dataset)
        self.train_size = len(self.train_dataloader.dataset)
        self.num_batches = len(self.test_dataloader)
        
        self.loss_function = loss_function
        self.optimizer = optimizer
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.num_epochs = num_epochs

        self.display_step_size = display_step_size
        self.evaluate = evaluate
        self.save_confusion_matrix = save_confusion_matrix
    
    def _get_prediction(self, X:torch.Tensor) -> torch.Tensor:
        '''
        This is the procedure to get a prediction for both training and testing. Will wrap it in this function quickly.

        params:
            - X (torch.Tensor): input data of size [batch_size, 1, 3, 224, 224]
        returns:
            - prediction (torch.Tensor): prediction from model of X
        '''
        # Squeeze:
        X = torch.squeeze(X)

        # Get prediction and set to device
        prediction = self.model(X)
        prediction = prediction.to(self.device)
        
        return prediction

    def _train_loop(self):
        '''
        Main train loop called by `self.train()`
        '''
        for batch_num, (X, y) in enumerate(self.train_dataloader):
            X, y = X.to(self.device), y.to(self.device)
            # Get prediction and calculate loss:
            prediction = self._get_prediction(X)
            loss = self.loss_function(prediction, y)

            # Backprop:
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

            # Show progress every display_step_size
            if batch_num % self.display_step_size == 0:
                loss, current = loss.item(), batch_num * len(X)
                print(f"Loss: {loss:>7f}  [{current:>5d}/{self.train_size:>5d}]")
    
    @torch.no_grad()
    def _test_loop(self):
        '''
        Main test loop called by `self.train()`.
        '''
        # Instantiate necessary variables
        test_loss, num_correct = 0, 0
        # Now we run evaluation loop:
        for X, y in self.test_dataloader:
            X, y = X.to(self.device), y.to(self.device)
            # Get prediction:
            prediction = self._get_prediction(X)
            # Get loss
            test_loss += self.loss_function(prediction, y).item()
            # If prediction is correct, add 1
            num_correct += (prediction.argmax(1) == y).type(torch.float).sum().item()
        
        # Get loss per batch and accuracy
        test_loss /= self.num_batches
        num_correct /= self.test_size
        print(f"Test Error: \n Accuracy: {(100*num_correct):>0.1f}%, Avg loss: {test_loss:>8f} \n")
    
    def train(self):
        # Main loop for training and testing:
        for e in range(self.num_epochs):
            print(f'Epoch {e} \n ---------------------')
            self._train_loop()
            self._test_loop()
        print('Training has been completed.')
        # Save model:
        model_root_dir = os.path.join(self.train_dataloader.dataset.data_dir, 'models')
        # If root directory doesn't exist, create it
        if not os.path.isdir(model_root_dir):
            os.mkdir(model_root_dir)
        # Create full path:
        model_path = os.path.join(model_root_dir, f'{self.model_name}_checkpoint_{e}.pth')
        # Save model:
        torch.save(self.model.state_dict(), model_path)
    
    @torch.no_grad()
    def evaluate(self, validation_dataloader:DataLoader):
        '''
        Run evaluation on validation set and return results.

        params:
            - validation_dataloader (Dataloader):
            - save_confusion_matrix (bool):
        '''
        all_gt = []
        all_predictions = []
        for X, y in validation_dataloader:
            # Set to device:
            X, y = X.to(self.device), y.to(self.device)
            # Get prediction probabilites:
            predictions = self.model(X)
            # Get labels:
            predictions_labels = predictions.argmax(1)
            # Save all values:
            all_gt.append(y.tolist())
            all_predictions(predictions_labels.tolist())
        
        # Flatten:
        all_gt = np.array(all_gt).flatten()
        all_predictions = np.array(all_predictions).flatten()
        
        # Get confusion matrix and visualize:
        cm = confusion_matrix(all_gt, all_predictions)
        disp = ConfusionMatrixDisplay(confusion_matrix=cm)
        disp.plot()
        plt.show()


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="vit4elm",
    version="1.1.0",
    author="Adam Sabra",
    description="Vision Transformers for Exotic Lattice Models",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=[
        'scikit-learn',
        'numpy',
        'matplotlib',
        'pandas',
        'transformers',
        'datasets',
        'torch',
        'torchvision',
        'pillow',
        'twine'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.8"
)